using System;
using System.Collections.Generic;
using System.Diagnostics;

using Microsoft.VisualStudio.OLE.Interop;

using OleConstants = Microsoft.VisualStudio.OLE.Interop.Constants;
using VSConstants = Microsoft.VisualStudio.VSConstants;

namespace Banquet
{
	internal class CommandHandler : IOleCommandTarget
	{
		private readonly Dictionary<Guid, CommandGroup> groups;
		private readonly List<CommandHandler> secondaryCommandHandlers;

		public CommandHandler()
		{
			this.groups = new Dictionary<Guid, CommandGroup>(2);
			this.secondaryCommandHandlers = new List<CommandHandler>();
		}

		private void AddCommand(Guid commandGroup, int commandId, StatusCallback statusCallback, ExecCallback execCallback)
		{
			Command command = new Command(commandGroup, commandId);
			command.StatusCallback = statusCallback;
			command.ExecCallback = execCallback;
			command.Supported = true;
			command.Enabled = false;
			command.Visible = true;

			CommandGroup group = GetCommandGroup(commandGroup);
			if (group == null)
			{
				group = new CommandGroup();
				this.groups.Add(commandGroup, group);
			}

			group.Commands.Add(commandId, command);
		}

		public void AddSecondaryCommandHandler(CommandHandler commandHandler)
		{
			Debug.Assert(commandHandler != null);
			this.secondaryCommandHandlers.Add(commandHandler);
		}

		protected bool DefaultQueryStatus(CommandHandler handler, Command command, int commandId)
		{
			command.Supported = true;
			command.Enabled = true;
			command.Visible = true;
			command.Checked = false;
			return true;
		}

		private CommandGroup GetCommandGroup(Guid commandGroup)
		{
			CommandGroup group;
			this.groups.TryGetValue(commandGroup, out group);
			return group;
		}

		protected virtual bool OnQueryStatus(Command command, int commandId)
		{
			return command.StatusCallback(this, command, commandId);
		}

		public void Register(Guid commandGroup, VSConstants.VSStd97CmdID commandId, StatusCallback statusCallback, ExecCallback execCallback)
		{
			Debug.Assert(commandGroup != Guid.Empty);
			Debug.Assert(statusCallback != null);
			Debug.Assert(execCallback != null);
			this.AddCommand(commandGroup, (int)commandId, statusCallback, execCallback);
		}

		public void Register(Guid commandGroup, int commandId, StatusCallback statusCallback, ExecCallback execCallback)
		{
			Debug.Assert(commandGroup != Guid.Empty);
			Debug.Assert(statusCallback != null);
			Debug.Assert(execCallback != null);
			this.AddCommand(commandGroup, commandId, statusCallback, execCallback);
		}

		public void RemoveSecondaryCommandHandler(CommandHandler commandHandler)
		{
			Debug.Assert(commandHandler != null);
			this.secondaryCommandHandlers.Remove(commandHandler);
		}

		#region IOleCommandTarget Members

		int IOleCommandTarget.Exec(ref Guid pguidCmdGroup, uint nCmdID, uint nCmdexecopt, IntPtr pvaIn, IntPtr pvaOut)
		{
			CommandGroup group = GetCommandGroup(pguidCmdGroup);
			if (group != null)
			{
				int commandId = (int)nCmdID;
				Command command = group.GetCommand(commandId);
				if (command != null)
				{
					if (!command.ExecCallback(this, command, commandId))
					{
						return (int)OleConstants.OLECMDERR_E_UNKNOWNGROUP;
					}
					return VSConstants.S_OK;
				}
			}

			foreach (IOleCommandTarget handler in this.secondaryCommandHandlers)
			{
				if (handler.Exec(ref pguidCmdGroup, nCmdID, nCmdexecopt, pvaIn, pvaOut) == VSConstants.S_OK)
				{
					return VSConstants.S_OK;
				}
			}

			return (int)OleConstants.OLECMDERR_E_UNKNOWNGROUP;
		}

		int IOleCommandTarget.QueryStatus(ref Guid pguidCmdGroup, uint cCmds, OLECMD[] prgCmds, IntPtr pCmdText)
		{
			Debug.Assert(cCmds == 1);
			Debug.Assert(prgCmds != null);

			CommandGroup group = GetCommandGroup(pguidCmdGroup);
			if (group != null)
			{
				int commandId = (int)prgCmds[0].cmdID;
				Command command = group.GetCommand(commandId);
				if (command != null)
				{
					command.OleCommandFlags = (OLECMDF)0;
					if (!OnQueryStatus(command, commandId))
					{
						return (int)OleConstants.OLECMDERR_E_UNKNOWNGROUP;
					}
					prgCmds[0].cmdf = (uint)command.OleCommandFlags;
					return VSConstants.S_OK;
				}
			}

			foreach (IOleCommandTarget handler in this.secondaryCommandHandlers)
			{
				if (handler.QueryStatus(ref pguidCmdGroup, cCmds, prgCmds, pCmdText) == VSConstants.S_OK)
				{
					return VSConstants.S_OK;
				}
			}

			return (int)OleConstants.OLECMDERR_E_UNKNOWNGROUP;
		}

		#endregion

		internal class CommandGroup
		{
			private readonly Dictionary<int, Command> commands;

			public CommandGroup()
			{
				this.commands = new Dictionary<int, Command>();
			}

			public IDictionary<int, Command> Commands
			{
				get
				{
					return this.commands;
				}
			}

			public Command GetCommand(int commandId)
			{
				Command command;
				this.commands.TryGetValue(commandId, out command);
				return command;
			}
		}

		internal class Command
		{
			private Guid m_group;
			private int m_id;

			private OLECMDF m_oleCommandFlags;

			private ExecCallback m_execCallback;
			private StatusCallback m_statusCallback;

			public Command(Guid commandGroup, int commandId)
			{
				this.m_group = commandGroup;
				this.m_id = commandId;
			}

			public bool Checked
			{
				get
				{
					return (this.m_oleCommandFlags & OLECMDF.OLECMDF_LATCHED) != 0;
				}
				set
				{
					if (value)
					{
						this.m_oleCommandFlags |= OLECMDF.OLECMDF_LATCHED;
					}
					else
					{
						this.m_oleCommandFlags &= ~OLECMDF.OLECMDF_LATCHED;
					}
				}
			}

			public bool Enabled
			{
				get
				{
					return (this.m_oleCommandFlags & OLECMDF.OLECMDF_ENABLED) != 0;
				}
				set
				{
					if (value)
					{
						this.m_oleCommandFlags |= OLECMDF.OLECMDF_ENABLED;
					}
					else
					{
						this.m_oleCommandFlags &= ~OLECMDF.OLECMDF_ENABLED;
					}
				}
			}

			public ExecCallback ExecCallback
			{
				get
				{
					return this.m_execCallback;
				}
				set
				{
					this.m_execCallback = value;
				}
			}

			public Guid Group
			{
				get
				{
					return this.m_group;
				}
			}

			public int Id
			{
				get
				{
					return this.m_id;
				}
			}

			public bool Latched
			{
				get
				{
					return this.Checked;
				}
				set
				{
					this.Checked = value;
				}
			}

			public OLECMDF OleCommandFlags
			{
				get
				{
					return this.m_oleCommandFlags;
				}
				set
				{
					this.m_oleCommandFlags = value;
				}
			}

			public StatusCallback StatusCallback
			{
				get
				{
					return this.m_statusCallback;
				}
				set
				{
					this.m_statusCallback = value;
				}
			}

			public bool Supported
			{
				get
				{
					return (this.m_oleCommandFlags & OLECMDF.OLECMDF_SUPPORTED) != 0;
				}
				set
				{
					if (value)
					{
						this.m_oleCommandFlags |= OLECMDF.OLECMDF_SUPPORTED;
					}
					else
					{
						this.m_oleCommandFlags &= ~OLECMDF.OLECMDF_SUPPORTED;
					}
				}
			}

			public bool Visible
			{
				get
				{
					return (this.m_oleCommandFlags & OLECMDF.OLECMDF_INVISIBLE) == 0;
				}
				set
				{
					if (value)
					{
						this.m_oleCommandFlags &= ~OLECMDF.OLECMDF_INVISIBLE;
					}
					else
					{
						this.m_oleCommandFlags |= OLECMDF.OLECMDF_INVISIBLE;
					}
				}
			}
		}

		public delegate bool ExecCallback(CommandHandler handler, CommandHandler.Command command, int commandId);

		public delegate bool StatusCallback(CommandHandler handler, CommandHandler.Command command, int commandId);
	}
}