using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Windows.Forms;

using Microsoft.VisualStudio.VirtualTreeGrid;

namespace Banquet
{
	internal class ExplorerTreeNode : IBranch
	{
		#region Fields

		private ExplorerTree tree;
		private ExplorerTreeNode parent;
		private ExplorerTreeNodeType type;
		private List<ExplorerTreeNode> children;
		private string name;
		private Uri location;

		#endregion

		#region Events

		public event BranchModificationEventHandler OnBranchModification;

		#endregion

		#region Constructors

		public ExplorerTreeNode(ExplorerTreeNodeType type, string name, Uri location, ExplorerTree tree)
		{
			this.tree = tree;
			this.type = type;
			this.children = new List<ExplorerTreeNode>();
			this.name = name;
			this.location = location;
		}

		public ExplorerTreeNode(ExplorerTreeNodeType type, string name, Uri location, ExplorerTreeNode parent)
			: this( type, name, location, parent.Tree)
		{
			this.parent = parent;
		}

		#endregion

		#region Methods

		public void Add(ExplorerTreeNode item)
		{
			int index = this.children.BinarySearch(item, ExplorerTreeNodeComparer.Instance);
			Debug.Assert(index < 0);
			Insert(~index, item);
		}

		public VirtualTreeLabelEditData BeginLabelEdit(int row, int column, VirtualTreeLabelEditActivationStyles activationStyle)
		{
			return VirtualTreeLabelEditData.Invalid;
		}

		public SubItemCellStyles ColumnStyles(int column)
		{
			return SubItemCellStyles.Simple;
		}

		public LabelEditResult CommitLabelEdit(int row, int column, string newText)
		{
			return LabelEditResult.CancelEdit;
		}

		public VirtualTreeAccessibilityData GetAccessibilityData(int row, int column)
		{
			return new VirtualTreeAccessibilityData();
		}

		public VirtualTreeDisplayData GetDisplayData(int row, int column, VirtualTreeDisplayDataMasks requiredData)
		{
			if (column != 0)
			{
				return VirtualTreeDisplayData.Empty;
			}

			VirtualTreeDisplayData data = new VirtualTreeDisplayData();

			if ((requiredData.Mask & VirtualTreeDisplayMasks.StateImage) != 0)
			{
				data.StateImageIndex = (short)this.children[row].GetStateImageIndex();
			}
			if ((requiredData.Mask & VirtualTreeDisplayMasks.Image) != 0)
			{
				bool expanded = (requiredData.StateMask & VirtualTreeDisplayStates.Expanded) != 0;
				data.Image = (short)this.children[row].GetImageIndex(expanded);
			}
			if ((requiredData.Mask & VirtualTreeDisplayMasks.SelectedImage) != 0)
			{
				bool expanded = (requiredData.StateMask & VirtualTreeDisplayStates.Expanded) != 0;
				data.SelectedImage = (short)this.children[row].GetSelectedImageIndex(expanded);
			}
			if ((requiredData.Mask & VirtualTreeDisplayMasks.State) != 0)
			{
				data.Bold = (this.children[row].Type == ExplorerTreeNodeType.Computer);
				//data.GrayText = (this.children[row].Type == ExplorerTreeNodeType.Directory);
			}
			//if ((requiredData.Mask & VirtualTreeDisplayMasks.ImageOverlays) != 0)
			//{
			//    data.OverlayIndex = 7;
			//}

			data.OverlayIndex = -1;
			return data;
		}

		public ExplorerTreeNodeImage GetImageIndex(bool expanded)
		{
			switch (this.type)
			{
			case ExplorerTreeNodeType.World:
				return ExplorerTreeNodeImage.World;
			case ExplorerTreeNodeType.Computer:
				return ExplorerTreeNodeImage.Computer;
			case ExplorerTreeNodeType.Repository:
				return ExplorerTreeNodeImage.Repository;
			case ExplorerTreeNodeType.Branch:
				return ExplorerTreeNodeImage.Branch;
			default:
				return expanded ? ExplorerTreeNodeImage.FolderOpen : ExplorerTreeNodeImage.FolderClosed;
			}
		}

		public object GetObject(int row, int column, ObjectStyle style, ref int options)
		{
			if (style == ObjectStyle.ExpandedBranch)
			{
				return this.children[row];
			}

			return null;
		}

		public ExplorerTreeNodeImage GetSelectedImageIndex(bool expanded)
		{
			return GetImageIndex(expanded);
		}

		public int GetStateImageIndex()
		{
			return -1;
		}

		public string GetText(int row, int column)
		{
			switch (column)
			{
			case 0:
				return this.children[row].Name ?? string.Empty;
			//case 1:
			//    return this.children[row].Location ?? string.Empty;
			//case 2:
			//    return this.children[row].Workspace ?? string.Empty;
			default:
				return null;
			}
		}

		public string GetTipText(int row, int column, ToolTipType tipType)
		{
			//return this.children[row].ToolTipText ?? string.Empty;
			return string.Empty;
		}

		public void Insert(int index, ExplorerTreeNode item)
		{
			this.children.Insert(index, item);

			BranchModificationEventHandler onBranchModification = OnBranchModification;
			if (onBranchModification != null)
			{
				onBranchModification(this, BranchModificationEventArgs.InsertItems(this, index - 1, 1));
			}
		}

		public bool IsExpandable(int row, int column)
		{
			if (column != 0)
			{
				return false;
			}

			return (this.children[row].Children.Count != 0);
		}

		public LocateObjectData LocateObject(object obj, ObjectStyle style, int locateOptions)
		{
			return new LocateObjectData();
		}

		public void OnDragEvent(object sender, int row, int column, DragEventType eventType, DragEventArgs args)
		{
		}

		public void OnGiveFeedback(GiveFeedbackEventArgs args, int row, int column)
		{
		}

		public void OnQueryContinueDrag(QueryContinueDragEventArgs args, int row, int column)
		{
		}

		public VirtualTreeStartDragData OnStartDrag(object sender, int row, int column, DragReason reason)
		{
			return VirtualTreeStartDragData.Empty;
		}

		public StateRefreshChanges SynchronizeState(int row, int column, IBranch matchBranch, int matchRow, int matchColumn)
		{
			return StateRefreshChanges.None;
		}

		public StateRefreshChanges ToggleState(int row, int column)
		{
			return StateRefreshChanges.None;
		}

		#endregion

		#region Properties

		public IList<ExplorerTreeNode> Children
		{
			get
			{
				return this.children;
			}
		}

		public BranchFeatures Features
		{
			get
			{
				return (BranchFeatures.DefaultPositionTracking | BranchFeatures.StateChanges | BranchFeatures.InsertsAndDeletes | BranchFeatures.Expansions);
			}
		}

		public Uri Location
		{
			get
			{
				return this.location;
			}
		}

		public string Name
		{
			get
			{
				return this.name;
			}
		}

		public ExplorerTreeNode Parent
		{
			get
			{
				return this.parent;
			}
		}

		//public virtual string ToolTipText
		//{
		//    get
		//    {
		//        return this.toolTipText;
		//    }
		//    set
		//    {
		//        this.toolTipText = value;
		//    }
		//}

		public ExplorerTree Tree
		{
			get
			{
				return this.tree;
			}
		}

		public ExplorerTreeNodeType Type
		{
			get
			{
				return this.type;
			}
		}

		public int UpdateCounter
		{
			get
			{
				//throw new NotImplementedException();
				return 0;
			}
		}

		public int VisibleItemCount
		{
			get
			{
				return this.children.Count;
			}
		}

		#endregion
	}
}