using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;

namespace Banquet
{
	internal class TreeComboBox : ComboBox
	{
		private static int DefaultIndentWidth = 8;
		private static int PaddingImageText = 2;

		private ImageList imageList;
		private bool drawFocusRect;

		public TreeComboBox()
		{
			base.DrawMode = DrawMode.OwnerDrawFixed;
		}

		public ImageList ImageList
		{
			get
			{
				return this.imageList;
			}
			set
			{
				this.imageList = value;
				if (value != null && base.ItemHeight < value.ImageSize.Height)
				{
					base.ItemHeight = value.ImageSize.Height;
				}
			}
		}

		protected virtual string GetComboEditText(TreeComboBoxItem item)
		{
			return item.Text;
		}

		protected override void OnPreviewKeyDown(PreviewKeyDownEventArgs e)
		{
			this.drawFocusRect = true;
			base.OnPreviewKeyDown(e);
		}

		protected override void OnDrawItem(DrawItemEventArgs e)
		{
			base.OnDrawItem(e);

			if (e.Index >= 0)
			{
				TreeComboBoxItem item = (TreeComboBoxItem)base.Items[e.Index];

				int indentWidth = (this.imageList == null) ? DefaultIndentWidth : (this.imageList.ImageSize.Width / 2);
				int x = item.IndentLevel * indentWidth + base.Padding.Left + 1;
				string comboEditText = item.Text;
				int y = 0;

				if ((e.State & DrawItemState.ComboBoxEdit) == DrawItemState.ComboBoxEdit)
				{
					x = PaddingImageText + base.Padding.Left + 1;
					comboEditText = GetComboEditText(item);
					y = PaddingImageText;
				}

				if (this.imageList != null)
				{
					x += this.ImageList.ImageSize.Width + PaddingImageText;
				}

				SizeF textSize = e.Graphics.MeasureString(comboEditText, this.Font);
				Rectangle rect = new Rectangle(x, e.Bounds.Y, Math.Min((int)textSize.Width, e.Bounds.Width - x + y) + 1, e.Bounds.Height);
				//Rectangle rect = new Rectangle(x, e.Bounds.Y, e.Bounds.Width - x + y, e.Bounds.Height);

				StringFormat format = new StringFormat();
				format.Alignment = StringAlignment.Near;
				format.Trimming = StringTrimming.EllipsisPath;
				format.FormatFlags = StringFormatFlags.NoWrap;

				Color backColor = this.BackColor;
				Color foreColor = this.ForeColor;
				if ((e.State & DrawItemState.Selected) == DrawItemState.Selected)
				{
					backColor = SystemColors.Highlight;
					foreColor = SystemColors.HighlightText;
				}

				using (SolidBrush brush = new SolidBrush(backColor))
				{
					e.Graphics.FillRectangle(brush, rect);
				}

				using (SolidBrush brush = new SolidBrush(foreColor))
				{
					e.Graphics.DrawString(comboEditText, this.Font, brush, rect, format);
				}

				if (this.drawFocusRect && (e.State & DrawItemState.Selected) == DrawItemState.Selected)
				{
					ControlPaint.DrawFocusRectangle(e.Graphics, rect);
				}

				if (this.imageList != null && item.ImageIndex >= 0)
				{
					e.Graphics.CompositingMode = CompositingMode.SourceOver;
					this.imageList.Draw(e.Graphics, x - this.imageList.ImageSize.Width - PaddingImageText, e.Bounds.Y, item.ImageIndex);
				}
			}
		}
	}
}