using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace Banquet
{
	internal static class ImageListHelper
	{
		private const string directoryKey = "|dir";
		private const string openModifierKey = "|open";
		private const string selectedModifierKey = "|selected";

		private static ImageList largeSystemImageList;
		private static ImageList smallSystemImageList;

		private static Dictionary<string, FileTypeInfo> indexMap = new Dictionary<string, FileTypeInfo>(StringComparer.OrdinalIgnoreCase);

		private static object @lock = new object();

		public static ImageList LargeSystemImageList
		{
			get
			{
				if (ImageListHelper.largeSystemImageList == null)
				{
					lock (ImageListHelper.@lock)
					{
						if (ImageListHelper.largeSystemImageList == null)
						{
							NativeMethods.ShFileInfoW psfi = new NativeMethods.ShFileInfoW();
							psfi.hIcon = IntPtr.Zero;
							psfi.iIcon = 0;
							psfi.dwAttributes = 0;
							psfi.szDisplayName = string.Empty;
							psfi.szTypeName = string.Empty;

							IntPtr hImageList = NativeMethods.SHGetFileInfoW(string.Empty, 0, ref psfi, Marshal.SizeOf(psfi), NativeMethods.ShgfiSysIconIndex | NativeMethods.ShgfiLargeIcon);
							ImageListHelper.largeSystemImageList = WrapImageList(hImageList);
						}
					}
				}

				return ImageListHelper.largeSystemImageList;
			}
		}

		public static ImageList SmallSystemImageList
		{
			get
			{
				if (ImageListHelper.smallSystemImageList == null)
				{
					lock (ImageListHelper.@lock)
					{
						if (ImageListHelper.smallSystemImageList == null)
						{
							NativeMethods.ShFileInfoW psfi = new NativeMethods.ShFileInfoW();
							psfi.hIcon = IntPtr.Zero;
							psfi.iIcon = 0;
							psfi.dwAttributes = 0;
							psfi.szDisplayName = string.Empty;
							psfi.szTypeName = string.Empty;

							IntPtr hImageList = NativeMethods.SHGetFileInfoW(string.Empty, 0, ref psfi, Marshal.SizeOf(psfi), NativeMethods.ShgfiSysIconIndex | NativeMethods.ShgfiSmallIcon);
							ImageListHelper.smallSystemImageList = WrapImageList(hImageList);
						}
					}
				}

				return ImageListHelper.smallSystemImageList;
			}
		}

		public static int GetSystemImageIndex(int size, string fileOrFolderName, int attributes)
		{
			string description;
			return GetSystemImageIndex(size, fileOrFolderName, attributes, false, false, out description);
		}

		public static int GetSystemImageIndex(int size, string fileOrFolderName, int attributes, bool openIcon, bool selectedIcon)
		{
			string description;
			return GetSystemImageIndex(size, fileOrFolderName, attributes, openIcon, selectedIcon, out description);
		}

		public static int GetSystemImageIndex(int imageSize, string fileOrFolderName, int attributes, bool openIcon, bool selectedIcon, out string description)
		{
			string key;

			description = string.Empty;

			int uFlags = imageSize | NativeMethods.ShgfiSysIconIndex | NativeMethods.ShgfiTypeName | NativeMethods.ShgfiUseFileAttributes;

			if ((attributes & (int)FileAttributes.Directory) != 0)
			{
				key = ImageListHelper.directoryKey;
				fileOrFolderName = string.Empty;
			}
			else if (fileOrFolderName != null)
			{
				key = Path.GetExtension(fileOrFolderName);
			}
			else
			{
				return -1;
			}

			if (openIcon)
			{
				uFlags |= NativeMethods.ShgfiOpenIcon;
				key = key + ImageListHelper.openModifierKey;
			}

			if (selectedIcon)
			{
				uFlags |= NativeMethods.ShgfiSelected;
				key = key + ImageListHelper.selectedModifierKey;
			}

			FileTypeInfo info = null;

			if (ImageListHelper.indexMap.TryGetValue(key, out info))
			{
				description = info.description;

				switch (imageSize)
				{
				case 0:
					if ((info.fetchedFlags & FileTypeInfoFetched.LargeIcon) != FileTypeInfoFetched.LargeIcon)
					{
						break;
					}
					return info.largeIconIndex;

				case 1:
					if ((info.fetchedFlags & FileTypeInfoFetched.SmallIcon) != FileTypeInfoFetched.SmallIcon)
					{
						break;
					}
					return info.smallIconIndex;
				}
			}

			NativeMethods.ShFileInfoW psfi = new NativeMethods.ShFileInfoW();
			psfi.hIcon = IntPtr.Zero;
			psfi.iIcon = 0;
			psfi.dwAttributes = 0;
			psfi.szDisplayName = string.Empty;
			psfi.szTypeName = string.Empty;

			if (NativeMethods.SHGetFileInfoW(fileOrFolderName, attributes, ref psfi, Marshal.SizeOf(psfi), uFlags) == IntPtr.Zero)
			{
				psfi.iIcon = -1;
			}
			else
			{
				if (psfi.hIcon != IntPtr.Zero)
				{
					NativeMethods.DestroyIcon(psfi.hIcon);
				}
				description = psfi.szTypeName;
			}

			if (info == null)
			{
				info = new FileTypeInfo();
			}

			info.description = description;

			if (imageSize == 1)
			{
				info.fetchedFlags |= FileTypeInfoFetched.SmallIcon;
				info.smallIconIndex = psfi.iIcon;
			}
			else
			{
				info.fetchedFlags |= FileTypeInfoFetched.LargeIcon;
				info.largeIconIndex = psfi.iIcon;
			}

			ImageListHelper.indexMap[key] = info;

			return psfi.iIcon;
		}

		private static ImageList WrapImageList(IntPtr hImageList)
		{
			int width;
			int height;

			ImageList list = new ImageList();

			Type type = typeof(ImageList);
			Type nativeType = type.GetNestedType("NativeImageList", BindingFlags.NonPublic);
			ConstructorInfo ctor = nativeType.GetConstructor(BindingFlags.NonPublic | BindingFlags.Instance, null, new Type[] { typeof(IntPtr) }, null);
			object instance = ctor.Invoke(new object[] { hImageList });

			FieldInfo field = type.GetField("nativeImageList", BindingFlags.NonPublic | BindingFlags.Instance);
			field.SetValue(list, instance);

			NativeMethods.ImageList_GetIconSize(hImageList, out width, out height);
			list.ImageSize = new Size(width, height);

			return list;
		}

		private class FileTypeInfo
		{
			public string description = string.Empty;
			public ImageListHelper.FileTypeInfoFetched fetchedFlags = ImageListHelper.FileTypeInfoFetched.None;
			public int largeIconIndex = -1;
			public int smallIconIndex = -1;
		}

		[Flags]
		private enum FileTypeInfoFetched
		{
			None,
			SmallIcon,
			LargeIcon
		}
	}
}