using System;
using System.Diagnostics;
using System.IO;

namespace Canonical.Bazaar
{
	internal static class FileSpec
	{
		public static string Canonicalize(string path)
		{
			Debug.Assert(!string.IsNullOrEmpty(path));
			Debug.Assert(string.Equals(path, Path.GetFullPath(path), StringComparison.Ordinal));

			string fullPath = Path.GetPathRoot(path);
			bool exists = Directory.Exists(fullPath);

			int j = path.IndexOf(Path.DirectorySeparatorChar) + 1;
			int i = path.IndexOf(Path.DirectorySeparatorChar, j);

			while (j > 0)
			{
				string s = path.Substring(j, (i < 0 ? path.Length : i) - j);
				j = i + 1;
				i = path.IndexOf(Path.DirectorySeparatorChar, j);

				if (!exists)
				{
					fullPath = Path.Combine(fullPath, s);
				}
				else
				{
					string[] entries = Directory.GetFileSystemEntries(fullPath, s);

					if (entries.Length == 0)
					{
						fullPath = Path.Combine(fullPath, s);
						exists = false;
					}
					else
					{
						Debug.Assert(entries.Length < 2);
						fullPath = Path.Combine(fullPath, entries[0]);
					}
				}
			}

			return fullPath;
		}

		public static string[] MakeAbsolute(string[] items, string folder)
		{
			Debug.Assert(items != null);

			string[] result = new string[items.Length];
			for (int i = 0; items.Length > i; i++)
			{
				Debug.Assert(!string.IsNullOrEmpty(items[i])); // TODO: throw exception if string.IsNullOrEmpty(items[i])
				result[i] = MakeAbsolute(items[i], folder);
			}
			return result;
		}

		public static string MakeAbsolute(string item, string folder)
		{
			Debug.Assert(!string.IsNullOrEmpty(item));
			Debug.Assert(!string.IsNullOrEmpty(folder));
			Debug.Assert(string.Equals(folder, Canonicalize(Path.GetFullPath(folder)), StringComparison.Ordinal));

			return Canonicalize(Path.GetFullPath(Path.Combine(folder, item)));
		}

		public static string[] MakeRelative(string[] items, string folder)
		{
			Debug.Assert(items != null);

			string[] result = new string[items.Length];
			for (int i = 0; i < items.Length; i++)
			{
				Debug.Assert(!string.IsNullOrEmpty(items[i])); // TODO: throw exception if string.IsNullOrEmpty(items[i])
				result[i] = MakeRelative(Canonicalize(Path.GetFullPath(items[i])), folder);
			}
			return result;
		}

		public static string MakeRelative(string item, string folder)
		{
			Debug.Assert(!string.IsNullOrEmpty(item));
			Debug.Assert(!string.IsNullOrEmpty(folder));
			Debug.Assert(string.Equals(folder, Canonicalize(Path.GetFullPath(folder)), StringComparison.Ordinal));

			if (folder.Length <= item.Length && string.Equals(folder, Canonicalize(item.Substring(0, folder.Length)), StringComparison.Ordinal))
			{
				if (item.Length == folder.Length)
				{
					return string.Empty;
				}
				if (0 < folder.Length && Path.DirectorySeparatorChar == folder[folder.Length - 1])
				{
					return item.Substring(folder.Length).Replace(Path.DirectorySeparatorChar, '/');
				}
				if (Path.DirectorySeparatorChar == item[folder.Length])
				{
					return item.Substring(folder.Length + 1).Replace(Path.DirectorySeparatorChar, '/');
				}
			}

			throw new PathNotChildException(item, folder);
		}
	}
}