using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;

using Python;

namespace Canonical.Bazaar
{
	public sealed class Repository : IDisposable // TODO: catch PythonExceptions and throw subclasses of BazaarException
	{
#if DEBUG
		static Repository() // HACK: for testing purposes only
		{
			string path = Environment.GetEnvironmentVariable("bzr.dev", EnvironmentVariableTarget.User);
			if (!string.IsNullOrEmpty(path))
			{
				using (PythonModule sys = PythonModule.Import("sys"))
				using (PythonList sys_path = sys.GetAttribute<PythonList>("path"))
				using (PythonString the_path = PythonString.CreateString(path))
				{
					sys_path.Append(the_path);
				}
			}
		}
#endif

		#region Fields

		private readonly PythonObject impl;
		private Uri location;

		#endregion

		#region Constructors

		internal Repository(PythonObject impl)
		{
			Debug.Assert(impl != null);
			this.impl = impl;
		}

		#endregion

		~Repository()
		{
			Dispose(false);
		}

		public void Close()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			if (disposing)
			{
				this.impl.Close();
			}
		}

		#region Properties

		public bool IsShared
		{
			get
			{
				try
				{
					using (PythonBool is_shared = this.impl.InvokeMethod<PythonBool>("is_shared"))
					{
						return is_shared.IsTrue;
					}
				}
				catch (PythonException e)
				{
					BazaarException.Handle(e);
					throw;
				}
			}
		}

		public Uri Location
		{
			get
			{
				if (this.location == null)
				{
					using (PythonObject bzrdir = this.impl.GetAttribute("bzrdir"))
					using (PythonObject root_transport = bzrdir.GetAttribute("root_transport"))
					using (PythonString @base = root_transport.GetAttribute<PythonString>("base"))
					{
						this.location = new Uri(@base.ToString());
					}
				}

				Debug.Assert(this.location != null);
				return this.location;
			}
		}

		#endregion

		#region Static Methods

		public static Repository Open(Uri location)
		{
			if (location == null)
			{
				throw new ArgumentNullException("location");
			}

			try
			{
				using (PythonString the_location = PythonString.CreateUnicode(location.AbsoluteUri))
				using (PythonModule bzrlib_repository = PythonModule.Import("bzrlib.repository"))
				using (PythonClass bzrlib_repository_Repository = bzrlib_repository.GetAttribute<PythonClass>("Repository"))
				{
					return new Repository(bzrlib_repository_Repository.InvokeMethod("open", the_location));
				}
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		public static bool IsLocalPathARepo(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}

			return File.Exists(Path.Combine(Path.Combine(Path.Combine(Path.GetFullPath(path), ".bzr"), "repository"), "format"));
		}

		#endregion
	}
}