using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;

using Python;

namespace Canonical.Bazaar
{
	public sealed class WorkingTree : IDisposable // TODO: catch PythonExceptions and throw subclasses of BazaarException
	{
#if DEBUG
		static WorkingTree() // HACK: for testing purposes only
		{
			string path = Environment.GetEnvironmentVariable("bzr.dev", EnvironmentVariableTarget.User);
			if (!string.IsNullOrEmpty(path))
			{
				using (PythonModule sys = PythonModule.Import("sys"))
				using (PythonList sys_path = sys.GetAttribute<PythonList>("path"))
				using (PythonString the_path = PythonString.CreateString(path))
				{
					sys_path.Append(the_path);
				}
			}
		}
#endif

		#region Fields

		private readonly PythonObject impl;
		private string basedir;
		private Uri location;

		#endregion

		#region Constructors

		internal WorkingTree(PythonObject impl)
		{
			Debug.Assert(impl != null);
			this.impl = impl;
		}

		#endregion

		~WorkingTree()
		{
			Dispose(false);
		}

		public void Close()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			if (disposing)
			{
				this.impl.Close();
			}
		}

		#region Properties

		public string BaseDirectory
		{
			get
			{
				if (this.basedir == null)
				{
					using (PythonString basedir = this.impl.GetAttribute<PythonString>("basedir"))
					{
						this.basedir = Path.GetFullPath(basedir.ToString());
						Debug.Assert(string.Equals(this.basedir, FileSpec.Canonicalize(this.basedir), StringComparison.Ordinal));
					}
				}

				Debug.Assert(this.basedir != null);
				return this.basedir;
			}
		}

		public Uri Location
		{
			get
			{
				if (this.location == null)
				{
					using (PythonObject bzrdir = this.impl.GetAttribute("bzrdir"))
					using (PythonObject root_transport = bzrdir.GetAttribute("root_transport"))
					using (PythonString @base = root_transport.GetAttribute<PythonString>("base"))
					{
						this.location = new Uri(@base.ToString());
					}
				}

				Debug.Assert(this.location != null);
				return this.location;
			}
		}

		#endregion

		#region Static Methods

		public static WorkingTree Open(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}

			path = FileSpec.Canonicalize(Path.GetFullPath(path));

			try
			{
				using (PythonString the_path = PythonString.CreateUnicode(path))
				using (PythonModule bzrlib_workingtree = PythonModule.Import("bzrlib.workingtree"))
				using (PythonClass bzrlib_workingtree_WorkingTree = bzrlib_workingtree.GetAttribute<PythonClass>("WorkingTree"))
				{
					return new WorkingTree(bzrlib_workingtree_WorkingTree.InvokeMethod("open", the_path));
				}
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		public static WorkingTree OpenContaining(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}

			path = FileSpec.Canonicalize(Path.GetFullPath(path));

			try
			{
				using (PythonString the_path = PythonString.CreateUnicode(path))
				using (PythonModule bzrlib_workingtree = PythonModule.Import("bzrlib.workingtree"))
				using (PythonClass bzrlib_workingtree_WorkingTree = bzrlib_workingtree.GetAttribute<PythonClass>("WorkingTree"))
				using (PythonTuple result = bzrlib_workingtree_WorkingTree.InvokeMethod<PythonTuple>("open_containing", the_path))
				{
					return new WorkingTree(result.GetItem(0));
				}
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		private static void Recurse(string path, string folder, IList<string> items)
		{
			Debug.Assert(!string.IsNullOrEmpty(path));
			Debug.Assert(items != null);

			foreach (string item in Directory.GetFiles(path))
			{
				Debug.Assert(string.Equals(item, FileSpec.Canonicalize(Path.GetFullPath(item)), StringComparison.Ordinal));
				items.Add(FileSpec.MakeRelative(item, folder));
			}

			foreach (string item in Directory.GetDirectories(path))
			{
				Debug.Assert(string.Equals(item, FileSpec.Canonicalize(Path.GetFullPath(item)), StringComparison.Ordinal));
				items.Add(FileSpec.MakeRelative(item, folder));
				Recurse(item, folder, items);
			}
		}

		#endregion

		#region Methods

		public int Add(string[] paths)
		{
			return Add(paths, false);
		}

		public int Add(string path)
		{
			return Add(new string[] { path }, false);
		}

		public int Add(string[] paths, bool isRecursive)
		{
			if (paths == null)
			{
				throw new ArgumentNullException("paths", Messages.ArgumentNull_Paths);
			}
			if (paths.Length == 0)
			{
				return 0;
			}

			/*
			string[] the_paths = new string[paths.Length];
			for (int i = 0; i < paths.Length; i++)
			{
				the_paths[i] = Path.GetFullPath(paths[i]); // TODO: check paths[i] is in basedir
			}

			using (PythonList file_list = PythonList.Create(the_paths))
			using (PythonBool recurse = PythonBool.Create(isRecursive))
			using (PythonNone action = PythonNone.Create())
			using (PythonBool save = PythonBool.Create(true))
			using (PythonModule bzrlib_add = PythonModule.Import("bzrlib.add"))
			using (PythonTuple result = bzrlib_add.InvokeMethod<PythonTuple>("smart_add_tree", this.impl, file_list, recurse, action, save))
			using (PythonInteger added = result.GetItem<PythonInteger>(0))
			{
				return added.ToInt32();
			}
			*/

			LockTreeWrite();

			try
			{
				List<string> items = new List<string>();

				for (int i = 0; i < paths.Length; i++)
				{
					string path = FileSpec.Canonicalize(Path.GetFullPath(paths[i]).TrimEnd(Path.DirectorySeparatorChar, Path.AltDirectorySeparatorChar));
					CollectUnversionedPaths(FileSpec.MakeRelative(path, BaseDirectory), items);
					if (isRecursive && Directory.Exists(path))
					{
						Recurse(path, BaseDirectory, items);
					}
				}

				if (items.Count == 0)
				{
					return 0;
				}

				using (PythonList files = PythonList.Create(items))
				using (this.impl.InvokeMethod("add", files))
				{
				}

				return items.Count;
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
			finally
			{
				Unlock();
			}
		}

		public int Add(string path, bool isRecursive)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			if (path.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "path");
			}

			return Add(new string[] { path }, isRecursive);
		}

		private void CollectUnversionedPaths(string path, IList<string> items)
		{
			Debug.Assert(!string.IsNullOrEmpty(path));
			Debug.Assert(path[path.Length - 1] != '/');
			Debug.Assert(items != null);

			if (!items.Contains(path))
			{
				using (PythonString the_path = PythonString.CreateUnicode(path))
				using (PythonObject id = this.impl.InvokeMethod("path2id", the_path))
				{
					if (id.IsFalse)
					{
						int i = path.LastIndexOf('/');
						if (i >= 0)
						{
							CollectUnversionedPaths(path.Substring(0, i), items);
						}
						items.Add(path);
					}
				}
			}
		}

		public void Commit(Change[] changes, string comment)
		{
			Commit(changes, null, comment);
		}

		public void Commit(Change[] changes, string author, string comment)
		{
			if (changes == null)
			{
				throw new ArgumentNullException("changes", Messages.ArgumentNull_Changes);
			}
			if (changes.Length == 0)
			{
				return;
			}

			string[] paths = FileSpec.MakeRelative(Change.ToAbsolutePaths(changes), BaseDirectory);

			try
			{
				using (PythonString message = PythonString.CreateUnicode(comment ?? string.Empty))
				using (PythonNone timestamp = PythonNone.Create())
				using (PythonNone timezone = PythonNone.Create())
				using (PythonObject committer = (author != null) ? (PythonObject)PythonString.CreateUnicode(author) : (PythonObject)PythonNone.Create())
				using (PythonList specific_files = PythonList.Create(paths))
				using (PythonNone rev_id = PythonNone.Create())
				using (PythonBool allow_pointless = PythonBool.Create(false))
				using (PythonBool strict = PythonBool.Create(false))
				using (PythonBool verbose = PythonBool.Create(false))
				using (PythonNone revprops = PythonNone.Create())
				using (PythonBool local = PythonBool.Create(false))
				using (PythonNone reporter = PythonNone.Create())
				using (PythonModule bzrlib_commit = PythonModule.Import("bzrlib.commit"))
				using (PythonClass bzrlib_commit_Commit = bzrlib_commit.GetAttribute<PythonClass>("Commit"))
				using (PythonObject commit = bzrlib_commit_Commit.Invoke())
				using (commit.InvokeMethod<PythonString>("commit", message, timestamp, timezone, committer, specific_files, rev_id, allow_pointless, strict, verbose, revprops, this.impl, local, reporter)) // TODO: return revision id
				{
				}
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		public void Delete(string[] paths)
		{
			if (paths == null)
			{
				throw new ArgumentNullException("paths", Messages.ArgumentNull_Paths);
			}
			if (paths.Length == 0)
			{
				return;
			}

			paths = FileSpec.MakeRelative(paths, BaseDirectory);

			try
			{
				using (PythonList files = PythonList.Create(paths))
				using (PythonBool verbose = PythonBool.Create(false))
				using (PythonNone to_file = PythonNone.Create())
				using (PythonBool keep_files = PythonBool.Create(false))
				using (PythonBool force = PythonBool.Create(true))
				using (this.impl.InvokeMethod("remove", files, verbose, to_file, keep_files, force))
				{
				}
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		public void Delete(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			if (path.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "path");
			}

			Delete(new string[] { path });
		}

		public Change[] GetChanges()
		{
			return GetChanges((string[])null, false, false);
		}

		public Change[] GetChanges(string[] paths)
		{
			return GetChanges(paths, false, false);
		}

		public Change[] GetChanges(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			if (path.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "path");
			}

			return GetChanges(new string[] { path }, false, false);
		}

		public Change[] GetChanges(bool includeUnchanged, bool includeUnknown)
		{
			return GetChanges((string[])null, includeUnchanged, includeUnknown);
		}

		public Change[] GetChanges(string[] paths, bool includeUnchanged, bool includeUnknown)
		{
			if (paths != null)
			{
				if (paths.Length == 0)
				{
					return new Change[] { };
				}

				paths = FileSpec.MakeRelative(paths, BaseDirectory);
			}

			List<Change> changes = new List<Change>();

			LockRead();

			try
			{
				using (PythonObject from_tree = this.impl.InvokeMethod("basis_tree"))
				using (PythonBool include_unchanged = PythonBool.Create(includeUnchanged))
				using (PythonObject specific_files = (paths != null) ? (PythonObject)PythonList.Create(paths) : (PythonObject)PythonNone.Create())
				using (PythonNone pb = PythonNone.Create())
				using (PythonNone extra_trees = PythonNone.Create())
				using (PythonBool require_versioned = PythonBool.Create(false))
				using (PythonBool want_unversioned = PythonBool.Create(includeUnknown))
				using (PythonIterable<PythonTuple> _iter_changes = this.impl.InvokeMethod<PythonIterable<PythonTuple>>("_iter_changes", from_tree, include_unchanged, specific_files, pb, extra_trees, require_versioned, want_unversioned))
				{
					foreach (PythonTuple change in _iter_changes)
					{
						changes.Add(new Change(this, change));
					}
				}

				return changes.ToArray();
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
			finally
			{
				Unlock();
			}
		}

		public Change[] GetChanges(string path, bool includeUnchanged, bool includeUnknown)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			if (path.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "path");
			}

			return GetChanges(new string[] { path }, includeUnchanged, includeUnknown);
		}

		public Conflict[] GetConflicts()
		{
			List<Conflict> conflicts = new List<Conflict>();

			try
			{
				using (PythonIterable<PythonObject> conflict_list = this.impl.InvokeMethod<PythonIterable<PythonObject>>("conflicts"))
				{
					foreach (PythonObject conflict in conflict_list)
					{
						using (PythonString path = conflict.GetAttribute<PythonString>("path"))
						using (PythonString typestring = conflict.GetAttribute<PythonString>("typestring"))
						{
							conflicts.Add(new Conflict(this, path.ToString(), typestring.ToString()));
						}
					}
				}

				return conflicts.ToArray();
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		public string GetAbsolutePath(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			if (path.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "path");
			}

			/*
			using (PythonString filename = PythonString.CreateUnicode(path))
			using (PythonString abspath = this.impl.InvokeMethod<PythonString>("abspath", filename))
			{
				return Path.GetFullPath(abspath.ToString());
			}
			*/

			return FileSpec.MakeAbsolute(path, BaseDirectory);
		}

		public string GetRelativePath(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			if (path.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "path");
			}

			/*
			using (PythonString the_path = PythonString.CreateUnicode(Path.GetFullPath(path)))
			using (PythonString relpath = this.impl.InvokeMethod<PythonString>("relpath", the_path))
			{
				return relpath.ToString();
			}
			*/

			return FileSpec.MakeRelative(FileSpec.Canonicalize(Path.GetFullPath(path)), BaseDirectory);
		}

		public bool IsVersioned(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			if (path.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "path");
			}

			path = FileSpec.MakeRelative(FileSpec.Canonicalize(Path.GetFullPath(path)), BaseDirectory);

			try
			{
				using (PythonString the_path = PythonString.CreateUnicode(path))
				using (PythonObject id = this.impl.InvokeMethod("path2id", the_path))
				{
					return id.IsTrue;
				}
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		internal void LockRead()
		{
			using (this.impl.InvokeMethod("lock_read"))
			{
			}
		}

		internal void LockTreeWrite()
		{
			using (this.impl.InvokeMethod("lock_tree_write"))
			{
			}
		}

		public void Merge(Uri location)
		{
			throw new NotImplementedException(); // TODO
		}

		public bool Rename(string oldPath, string newPath)
		{
			return Rename(oldPath, newPath, false);
		}

		public bool Rename(string oldPath, string newPath, bool isAfter)
		{
			if (oldPath == null)
			{
				throw new ArgumentNullException("oldPath", Messages.ArgumentNull_Path);
			}
			if (oldPath.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "oldPath");
			}
			if (newPath == null)
			{
				throw new ArgumentNullException("newPath", Messages.ArgumentNull_Path);
			}
			if (newPath.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "newPath");
			}

			oldPath = Path.GetFullPath(oldPath);
			newPath = Path.GetFullPath(newPath);

			if (isAfter)
			{
				newPath = FileSpec.Canonicalize(newPath);
			}
			else
			{
				oldPath = FileSpec.Canonicalize(oldPath);
			}

			oldPath = FileSpec.MakeRelative(oldPath, BaseDirectory);
			newPath = FileSpec.MakeRelative(newPath, BaseDirectory);

			try
			{
				using (PythonString from_rel = PythonString.CreateUnicode(oldPath))
				using (PythonObject id = this.impl.InvokeMethod("path2id", from_rel))
				{
					if (id.IsFalse)
					{
						return false;
					}
					else
					{
						using (PythonString to_rel = PythonString.CreateUnicode(newPath))
						using (PythonBool after = PythonBool.Create(isAfter))
						using (this.impl.InvokeMethod("rename_one", from_rel, to_rel, after))
						{
							return true;
						}
					}
				}
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		public void ResolveConflict(Conflict conflict)
		{
			if (conflict == null)
			{
				throw new ArgumentNullException("conflict", Messages.ArgumentNull_Conflict);
			}
			// TODO: check conflict belongs to workingtree

			try
			{
				using (PythonString path = PythonString.CreateUnicode(null)) // TODO: extract path from conflict
				using (PythonList paths = PythonList.Create(path))
				using (PythonBool ignore_misses = PythonBool.Create(false))
				using (PythonModule bzrlib_conflicts = PythonModule.Import("bzrlib.conflicts"))
				using (bzrlib_conflicts.InvokeMethod("resolve", this.impl, paths, ignore_misses))
				{
				}
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		public void Revert(string[] paths)
		{
			if (paths == null)
			{
				throw new ArgumentNullException("paths", Messages.ArgumentNull_Paths);
			}
			if (paths.Length == 0)
			{
				return;
			}

			paths = FileSpec.MakeRelative(paths, BaseDirectory);

			try
			{
				using (PythonList filenames = PythonList.Create(paths))
				using (PythonNone old_tree = PythonNone.Create())
				using (PythonBool backups = PythonBool.Create(false))
				using (this.impl.InvokeMethod<PythonList>("revert", filenames, old_tree, backups)) // TODO: return conflicts
				{
				}
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		public void Revert(string path)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			if (path.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "path");
			}

			Revert(new string[] { path });
		}

		public void Revert(string[] paths, VersionSpec revision)
		{
			if (paths == null)
			{
				throw new ArgumentNullException("paths", Messages.ArgumentNull_Paths);
			}
			if (revision == null)
			{
				throw new ArgumentNullException("revision", Messages.ArgumentNull_VersionSpec);
			}
			if (paths.Length == 0)
			{
				return;
			}

			paths = FileSpec.MakeRelative(paths, BaseDirectory);

			try
			{
				using (PythonList filenames = PythonList.Create(paths))
				using (PythonString rev_id = PythonString.CreateString(null)) // TODO: get rev_id from versionspec
				using (PythonObject branch = this.impl.GetAttribute("branch"))
				using (PythonObject repository = branch.GetAttribute("repository"))
				using (PythonObject old_tree = repository.InvokeMethod("revision_tree", rev_id))
				using (PythonBool backups = PythonBool.Create(false))
				using (this.impl.InvokeMethod<PythonList>("revert", filenames, old_tree, backups)) // TODO: return conflicts
				{
				}
			}
			catch (PythonException e)
			{
				BazaarException.Handle(e);
				throw;
			}
		}

		public void Revert(string path, VersionSpec revision)
		{
			if (path == null)
			{
				throw new ArgumentNullException("path", Messages.ArgumentNull_Path);
			}
			if (path.Length == 0)
			{
				throw new ArgumentException(Messages.Argument_EmptyPath, "path");
			}

			Revert(new string[] { path }, revision);
		}

		public void Revert(Change change)
		{
			if (change == null)
			{
				throw new ArgumentNullException("change", Messages.ArgumentNull_Change);
			}

			Revert(new string[] { change.FullPath }); // TODO: check change belongs to workingtree
		}

		public void Revert(Change[] changes)
		{
			if (changes == null)
			{
				throw new ArgumentNullException("changes", Messages.ArgumentNull_Changes);
			}
			if (changes.Length == 0)
			{
				return;
			}

			Revert(Change.ToAbsolutePaths(changes)); // TODO: check changes belong to workingtree
		}

		internal void Unlock()
		{
			using (this.impl.InvokeMethod("unlock"))
			{
			}
		}

		public void Update()
		{
			using (this.impl.InvokeMethod<PythonInteger>("update")) // TODO: return result
			{
			}
		}

		#endregion
	}
}