using System;
using System.Diagnostics;
using System.Threading;

namespace Python
{
	public static class PythonEngine
	{
		#region Fields

		private static readonly GIL gil = new GIL();

		#endregion

		#region Properties

		public static string BuildInfo
		{
			get
			{
				return NativeMethods.Py_GetBuildInfo();
			}
		}

		public static string Compiler
		{
			get
			{
				return NativeMethods.Py_GetCompiler();
			}
		}

		public static string Copyright
		{
			get
			{
				string s = NativeMethods.Py_GetCopyright();

				if (Environment.NewLine != "\n")
				{
					s = s.Replace("\n", Environment.NewLine);
				}

				return s;
			}
		}

		public static IDisposable GlobalInterpreterLock
		{
			get
			{
				gil.Create();
				return gil;
			}
		}

		public static string Platform
		{
			get
			{
				return NativeMethods.Py_GetPlatform();
			}
		}

		public static string ProgramName
		{
			get
			{
				return NativeMethods.Py_GetProgramName();
			}
			set
			{
				NativeMethods.Py_SetProgramName(value);
			}
		}

		public static string PythonHome
		{
			get
			{
				return NativeMethods.Py_GetPythonHome();
			}
			set
			{
				NativeMethods.Py_SetPythonHome(value);
			}
		}

		public static string Version
		{
			get
			{
				return NativeMethods.Py_GetVersion();
			}
		}

		#endregion

		#region Methods

		[Conditional("DEBUG")]
		public static void AssertGlobalInterpreterLock()
		{
			gil.Assert();
		}

		public static void EnsureGlobalInterpreterLock()
		{
			gil.Assert();
		}

		#endregion

		private class GIL : IDisposable
		{
			private long counter;
			private IntPtr token;

			public void Assert()
			{
				if (Interlocked.Read(ref this.counter) == 0)
				{
					throw new InvalidOperationException();
				}
			}

			public void Create()
			{
				Monitor.Enter(this);

				if (Interlocked.Increment(ref this.counter) == 1)
				{
					this.token = NativeMethods.PyGILState_Ensure();
				}
			}

			public void Dispose()
			{
				if (Interlocked.Decrement(ref this.counter) == 0)
				{
					NativeMethods.PyGILState_Release(this.token);
				}

				Monitor.Exit(this);
			}
		}
	}
}