using System;

namespace Python
{
	public sealed class PythonException : Exception, IDisposable
	{
		#region Fields

		private readonly PythonType type;
		private readonly PythonObject value;
		private readonly PythonObject traceback;

		private string message;
		private string stackTrace;

		#endregion

		#region Constructors

		private PythonException()
		{
			this.type = new PythonType();
			this.value = new PythonObject();
			this.traceback = new PythonObject();

			NativeMethods.PyErr_Fetch(ref this.type, ref this.value, ref this.traceback);

			this.message = this.value.ToString();

			if (!this.type.IsInvalid)
			{
				this.message = string.Concat(this.type.Name, ": ", this.message);
			}
		}

		#endregion

		~PythonException()
		{
			Dispose(false);
		}

		public void Dispose()
		{
			Dispose(true);
		}

		private void Dispose(bool disposing)
		{
			if (disposing)
			{
				this.type.Dispose();
				this.value.Dispose();
				this.traceback.Dispose();
			}
		}

		#region Properties

		public override string Message
		{
			get
			{
				return this.message;
			}
		}

		public override string StackTrace
		{
			get
			{
				if (this.stackTrace == null)
				{
					if (this.traceback.IsInvalid)
					{
						this.stackTrace = string.Empty;
					}
					else
					{
						using (PythonModule tb_module = PythonModule.Import("traceback"))
						using (PythonList tb_result = tb_module.InvokeMethod<PythonList>("format_tb", this.traceback))
						{
							tb_result.Reverse();

							using (PythonString newline = PythonString.CreateString(string.Empty))
							using (PythonString result = newline.Join(tb_result))
							{
								this.stackTrace = result.ToString();

								if (Environment.NewLine != "\n")
								{
									this.stackTrace = this.stackTrace.Replace("\n", Environment.NewLine);
								}
							}
						}
					}

					this.stackTrace += base.StackTrace;
				}

				return this.stackTrace;
			}
		}

		public PythonType Type
		{
			get
			{
				return this.type;
			}
		}

		public PythonObject Value
		{
			get
			{
				return this.value;
			}
		}

		#endregion

		#region Static Methods

		internal static T Clear<T>(T o, T @default) where T : PythonObject
		{
			if (o.IsInvalid)
			{
				NativeMethods.PyErr_Clear();
				return @default;
			}

			return o;
		}

		internal static void ThrowIfErrorOccurred()
		{
			if (NativeMethods.PyErr_Occurred())
			{
				throw new PythonException();
			}
		}

		internal static T ThrowIfInvalid<T>(T o) where T : PythonObject
		{
			if (o.IsInvalid)
			{
				throw new PythonException();
			}

			return o;
		}

		internal static int ThrowIfNegative(int i)
		{
			if (i < 0)
			{
				throw new PythonException();
			}

			return i;
		}

		#endregion
	}
}