using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace Python
{
	public class PythonObject : SafeHandle, IEquatable<PythonObject>, IComparable<PythonObject>, IComparable, IDisposable
	{
		private static readonly Dictionary<IntPtr, PythonType> types = new Dictionary<IntPtr, PythonType>();

		public static readonly PythonType BooleanType = GetBuiltinType("BooleanType");
		public static readonly PythonType NoneType = GetBuiltinType("NoneType");
		public static readonly PythonType StringType = GetBuiltinType("StringType");
		public static readonly PythonType UnicodeType = GetBuiltinType("UnicodeType");

		private static PythonType GetBuiltinType(string name)
		{
			using (PythonModule module = PythonException.ThrowIfInvalid(NativeMethods.PyImport_ImportModule("types")))
			{
				PythonType type = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsType(module, name));
				types.Add(type.handle, type);
				return type;
			}
		}

		public PythonObject()
			: base(IntPtr.Zero, true)
		{
		}

		public sealed override bool IsInvalid
		{
			get
			{
				return IsClosed || (base.handle == IntPtr.Zero);
			}
		}

		protected internal unsafe IntPtr NewReference()
		{
			if (IsInvalid)
			{
				throw new InvalidOperationException();
			}

			NativeMethods.PyObject* op = (NativeMethods.PyObject*)base.handle;
			op->ob_refcnt = (IntPtr)((byte*)op->ob_refcnt + 1);
			return (IntPtr)op;
		}

		protected sealed override unsafe bool ReleaseHandle()
		{
			NativeMethods.PyObject* op = (NativeMethods.PyObject*)base.handle;

			IntPtr refcnt = (IntPtr)((byte*)op->ob_refcnt - 1);
			if (refcnt != IntPtr.Zero)
			{
				op->ob_refcnt = refcnt;
			}
			else
			{
				NativeMethods.Py_DecRef(op);
			}
			return true;
		}

		internal unsafe void SetHandle(NativeMethods.PyObject* op)
		{
			base.SetHandle((IntPtr)op);
		}

		public unsafe PythonType Type
		{
			get
			{
				NativeMethods.PyObject* op = (NativeMethods.PyObject*)base.handle;
				IntPtr handle = (IntPtr)(op->ob_type);
				PythonType type;

				lock (types)
				{
					if (!types.TryGetValue(handle, out type))
					{
						type = PythonException.ThrowIfInvalid(NativeMethods.PyObject_Type(this));
						types.Add(type.handle, type);
					}
				}

				return type;
			}
		}

		public bool HasAttribute(string name)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			return NativeMethods.PyObject_HasAttrString(this, name) > 0;
		}

		public bool HasAttribute(PythonObject name)
		{
			if ((object)name == null)
			{
				throw new ArgumentNullException("name");
			}

			return NativeMethods.PyObject_HasAttr(this, name) > 0;
		}

		public PythonObject GetAttribute(string name)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrString(this, name));
		}

		public T GetAttribute<T>(string name) where T : PythonObject, new()
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrString<T>(this, name));
		}

		public PythonObject GetAttribute(string name, PythonObject @default)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.Clear(NativeMethods.PyObject_GetAttrString(this, name), @default);
		}

		public T GetAttribute<T>(string name, T @default) where T : PythonObject, new()
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.Clear(NativeMethods.PyObject_GetAttrString<T>(this, name), @default);
		}

		public PythonObject GetAttribute(PythonObject name)
		{
			if ((object)name == null)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttr(this, name));
		}

		public T GetAttribute<T>(PythonObject name) where T : PythonObject, new()
		{
			if ((object)name == null)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttr<T>(this, name));
		}

		public PythonObject GetAttribute(PythonObject name, PythonObject @default)
		{
			if ((object)name == null)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.Clear(NativeMethods.PyObject_GetAttr(this, name), @default);
		}

		public T GetAttribute<T>(PythonObject name, T @default) where T : PythonObject, new()
		{
			if ((object)name == null)
			{
				throw new ArgumentNullException("name");
			}

			return PythonException.Clear(NativeMethods.PyObject_GetAttr<T>(this, name), @default);
		}

		public void SetAttribute(string name, PythonObject value)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if (value == null)
			{
				throw new ArgumentNullException("value");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_SetAttrString(this, name, value));
		}

		public void SetAttribute(PythonObject name, PythonObject value)
		{
			if ((object)name == null)
			{
				throw new ArgumentNullException("name");
			}
			if ((object)value == null)
			{
				throw new ArgumentNullException("value");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_SetAttr(this, name, value));
		}

		public void DeleteAttribute(string name)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_DelAttrString(this, name));
		}

		public void DeleteAttribute(PythonObject name)
		{
			if ((object)name == null)
			{
				throw new ArgumentNullException("name");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_DelAttr(this, name));
		}

		public PythonObject GetItem(PythonObject key)
		{
			if ((object)key == null)
			{
				throw new ArgumentNullException("key");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem(this, key));
		}

		public T GetItem<T>(PythonObject key) where T : PythonObject, new()
		{
			if ((object)key == null)
			{
				throw new ArgumentNullException("key");
			}

			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem<T>(this, key));
		}

		public PythonObject GetItem(string key)
		{
			if (key == null)
			{
				throw new ArgumentNullException("key");
			}

			using (PythonString k = PythonException.ThrowIfInvalid(NativeMethods.PyString_FromStringAndSize(key, key.Length)))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem(this, k));
			}
		}

		public T GetItem<T>(string key) where T : PythonObject, new()
		{
			if (key == null)
			{
				throw new ArgumentNullException("key");
			}

			using (PythonString k = PythonException.ThrowIfInvalid(NativeMethods.PyString_FromStringAndSize(key, key.Length)))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem<T>(this, k));
			}
		}

		public PythonObject GetItem(int index)
		{
			using (PythonInteger k = PythonException.ThrowIfInvalid(NativeMethods.PyInt_FromInt32(index)))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem(this, k));
			}
		}

		public T GetItem<T>(int index) where T : PythonObject, new()
		{
			using (PythonInteger k = PythonException.ThrowIfInvalid(NativeMethods.PyInt_FromInt32(index)))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetItem<T>(this, k));
			}
		}

		public void SetItem(PythonObject key, PythonObject value)
		{
			if ((object)key == null)
			{
				throw new ArgumentNullException("key");
			}
			if ((object)value == null)
			{
				throw new ArgumentNullException("value");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_SetItem(this, key, value));
		}

		public void SetItem(string key, PythonObject value)
		{
			if (key == null)
			{
				throw new ArgumentNullException("key");
			}
			if (value == null)
			{
				throw new ArgumentNullException("value");
			}

			using (PythonString k = PythonException.ThrowIfInvalid(NativeMethods.PyString_FromStringAndSize(key, key.Length)))
			{
				PythonException.ThrowIfNegative(NativeMethods.PyObject_SetItem(this, k, value));
			}
		}

		public void SetItem(int index, PythonObject value)
		{
			if ((object)value == null)
			{
				throw new ArgumentNullException("value");
			}

			using (PythonInteger k = PythonException.ThrowIfInvalid(NativeMethods.PyInt_FromInt32(index)))
			{
				PythonException.ThrowIfNegative(NativeMethods.PyObject_SetItem(this, k, value));
			}
		}

		public void DeleteItem(PythonObject key)
		{
			if ((object)key == null)
			{
				throw new ArgumentNullException("key");
			}

			PythonException.ThrowIfNegative(NativeMethods.PyObject_DelItem(this, key));
		}

		public void DeleteItem(string key)
		{
			if ((object)key == null)
			{
				throw new ArgumentNullException("key");
			}

			using (PythonString k = PythonException.ThrowIfInvalid(NativeMethods.PyString_FromStringAndSize(key, key.Length)))
			{
				PythonException.ThrowIfNegative(NativeMethods.PyObject_DelItem(this, k));
			}
		}

		public void DeleteItem(int index)
		{
			using (PythonInteger k = PythonException.ThrowIfInvalid(NativeMethods.PyInt_FromInt32(index)))
			{
				PythonException.ThrowIfNegative(NativeMethods.PyObject_DelItem(this, k));
			}
		}

		public PythonIterator<T> GetIterator<T>() where T : PythonObject, new()
		{
			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetIter<T>(this));
		}

		public PythonObject Invoke()
		{
			using (PythonTuple a = PythonTuple.Create())
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(this, a));
			}
		}

		public T Invoke<T>() where T : PythonObject, new()
		{
			using (PythonTuple a = PythonTuple.Create())
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(this, a));
			}
		}

		public PythonObject Invoke(PythonObject arg0)
		{
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}

			using (PythonTuple a = PythonTuple.Create(arg0))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(this, a));
			}
		}

		public T Invoke<T>(PythonObject arg0) where T : PythonObject, new()
		{
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}

			using (PythonTuple a = PythonTuple.Create(arg0))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(this, a));
			}
		}

		public PythonObject Invoke(PythonObject arg0, PythonObject arg1)
		{
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}
			if ((object)arg1 == null)
			{
				throw new ArgumentNullException("arg1");
			}

			using (PythonTuple a = PythonTuple.Create(arg0, arg1))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(this, a));
			}
		}

		public T Invoke<T>(PythonObject arg0, PythonObject arg1) where T : PythonObject, new()
		{
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}
			if ((object)arg1 == null)
			{
				throw new ArgumentNullException("arg1");
			}

			using (PythonTuple a = PythonTuple.Create(arg0, arg1))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(this, a));
			}
		}

		public PythonObject Invoke(PythonObject arg0, PythonObject arg1, PythonObject arg2)
		{
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}
			if ((object)arg1 == null)
			{
				throw new ArgumentNullException("arg1");
			}
			if ((object)arg2 == null)
			{
				throw new ArgumentNullException("arg2");
			}

			using (PythonTuple a = PythonTuple.Create(arg0, arg1, arg2))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(this, a));
			}
		}

		public T Invoke<T>(PythonObject arg0, PythonObject arg1, PythonObject arg2) where T : PythonObject, new()
		{
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}
			if ((object)arg1 == null)
			{
				throw new ArgumentNullException("arg1");
			}
			if ((object)arg2 == null)
			{
				throw new ArgumentNullException("arg2");
			}

			using (PythonTuple a = PythonTuple.Create(arg0, arg1, arg2))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(this, a));
			}
		}

		public PythonObject Invoke(params PythonObject[] args)
		{
			if (args == null)
			{
				throw new ArgumentNullException("args");
			}

			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(this, a));
			}
		}

		public T Invoke<T>(params PythonObject[] args) where T : PythonObject, new()
		{
			if (args == null)
			{
				throw new ArgumentNullException("args");
			}

			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(this, a));
			}
		}

		public PythonObject Invoke(PythonObject[] args, PythonDictionary kw)
		{
			if (args == null)
			{
				throw new ArgumentNullException("args");
			}
			if ((object)kw == null)
			{
				throw new ArgumentNullException("kw");
			}

			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_Call(this, a, kw));
			}
		}

		public T Invoke<T>(PythonObject[] args, PythonDictionary kw) where T : PythonObject, new()
		{
			if (args == null)
			{
				throw new ArgumentNullException("args");
			}
			if ((object)kw == null)
			{
				throw new ArgumentNullException("kw");
			}

			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_Call<T>(this, a, kw));
			}
		}

		public PythonObject InvokeMethod(string name)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create())
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(m, a));
			}
		}

		public T InvokeMethod<T>(string name) where T : PythonObject, new()
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create())
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(m, a));
			}
		}

		public PythonObject InvokeMethod(string name, PythonObject arg0)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(m, a));
			}
		}

		public T InvokeMethod<T>(string name, PythonObject arg0) where T : PythonObject, new()
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(m, a));
			}
		}

		public PythonObject InvokeMethod(string name, PythonObject arg0, PythonObject arg1)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}
			if ((object)arg1 == null)
			{
				throw new ArgumentNullException("arg1");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0, arg1))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(m, a));
			}
		}

		public T InvokeMethod<T>(string name, PythonObject arg0, PythonObject arg1) where T : PythonObject, new()
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}
			if ((object)arg1 == null)
			{
				throw new ArgumentNullException("arg1");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0, arg1))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(m, a));
			}
		}

		public PythonObject InvokeMethod(string name, PythonObject arg0, PythonObject arg1, PythonObject arg2)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}
			if ((object)arg1 == null)
			{
				throw new ArgumentNullException("arg1");
			}
			if ((object)arg2 == null)
			{
				throw new ArgumentNullException("arg2");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0, arg1, arg2))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(m, a));
			}
		}

		public T InvokeMethod<T>(string name, PythonObject arg0, PythonObject arg1, PythonObject arg2) where T : PythonObject, new()
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if ((object)arg0 == null)
			{
				throw new ArgumentNullException("arg0");
			}
			if ((object)arg1 == null)
			{
				throw new ArgumentNullException("arg1");
			}
			if ((object)arg2 == null)
			{
				throw new ArgumentNullException("arg2");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(arg0, arg1, arg2))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(m, a));
			}
		}

		public PythonObject InvokeMethod(string name, params PythonObject[] args)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if (args == null)
			{
				throw new ArgumentNullException("args");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject(m, a));
			}
		}

		public T InvokeMethod<T>(string name, params PythonObject[] args) where T : PythonObject, new()
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if (args == null)
			{
				throw new ArgumentNullException("args");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_CallObject<T>(m, a));
			}
		}

		public PythonObject InvokeMethod(string name, PythonObject[] args, PythonDictionary kw)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if (args == null)
			{
				throw new ArgumentNullException("args");
			}
			if ((object)kw == null)
			{
				throw new ArgumentNullException("kw");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_Call(m, a, kw));
			}
		}

		public T InvokeMethod<T>(string name, PythonObject[] args, PythonDictionary kw) where T : PythonObject, new()
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}
			if (args == null)
			{
				throw new ArgumentNullException("args");
			}
			if ((object)kw == null)
			{
				throw new ArgumentNullException("kw");
			}

			using (PythonMethod m = PythonException.ThrowIfInvalid(NativeMethods.PyObject_GetAttrStringAsMethod(this, name)))
			using (PythonTuple a = PythonTuple.Create(args))
			{
				return PythonException.ThrowIfInvalid(NativeMethods.PyObject_Call<T>(m, a, kw));
			}
		}

		public bool IsInstanceOf(PythonType type)
		{
			if ((object)type == null)
			{
				throw new ArgumentNullException("type");
			}

			return type.IsSubtypeOf(this.Type);
		}

		public bool IsCallable
		{
			get
			{
				return NativeMethods.PyCallable_Check(this) > 0;
			}
		}

		public bool IsTrue
		{
			get
			{
				return PythonException.ThrowIfNegative(NativeMethods.PyObject_IsTrue(this)) > 0;
			}
		}

		public bool IsFalse
		{
			get
			{
				return PythonException.ThrowIfNegative(NativeMethods.PyObject_Not(this)) > 0;
			}
		}

		public PythonList Dir()
		{
			return PythonException.ThrowIfInvalid(NativeMethods.PyObject_Dir(this));
		}

		public string Repr()
		{
			using (PythonString repr = PythonException.ThrowIfInvalid(NativeMethods.PyObject_Repr(this)))
			{
				return repr.ToString();
			}
		}

		public override string ToString()
		{
			using (PythonString unicode = PythonException.ThrowIfInvalid(NativeMethods.PyObject_Unicode(this)))
			{
				return unicode.ToString();
			}
		}

		#region IEquatable Members

		public override int GetHashCode()
		{
			using (PythonInteger o = PythonException.ThrowIfInvalid(NativeMethods.PyObject_Hash(this)))
			{
				return o.ToInt32();
			}
		}

		public override bool Equals(object obj)
		{
			return CompareTo(obj as PythonObject) == 0;
		}

		public bool Equals(PythonObject other)
		{
			return CompareTo(other) == 0;
		}

		public static bool Equals(PythonObject objA, PythonObject objB)
		{
			if ((object)objA == null)
			{
				return ((object)objB == null);
			}

			return objA.CompareTo(objB) == 0;
		}

		public static bool ReferenceEquals(PythonObject objA, PythonObject objB)
		{
			if ((object)objA == null || objA.IsInvalid)
			{
				return ((object)objB == null || objB.IsInvalid);
			}

			if ((object)objB == null || objB.IsInvalid)
			{
				return false;
			}

			return (objA.handle == objB.handle);
		}

		#endregion

		#region IComparable Members

		public int CompareTo(object obj)
		{
			return CompareTo(obj as PythonObject);
		}

		public int CompareTo(PythonObject other)
		{
			if ((object)other == null)
			{
				return 1;
			}

			if (ReferenceEquals(this, other))
			{
				return 0;
			}

			int value = NativeMethods.PyObject_Compare(this, other);
			PythonException.ThrowIfErrorOccurred();
			return value;
		}

		#endregion

		#region Operators

		public static bool operator ==(PythonObject objA, PythonObject objB)
		{
			return ReferenceEquals(objA, objB);
		}

		public static bool operator !=(PythonObject objA, PythonObject objB)
		{
			return !ReferenceEquals(objA, objB);
		}

		#endregion
	}
}